/* automatically generated by rust-bindgen 0.72.1 */

pub const MARKER_TYPE_SAMPLE_START: u32 = 0;
pub const MARKER_TYPE_SAMPLE_END: u32 = 1;
pub const MARKER_TYPE_BENCHMARK_START: u32 = 2;
pub const MARKER_TYPE_BENCHMARK_END: u32 = 3;
pub type InstrumentHooks = *mut u64;
extern "C" {
    pub fn instrument_hooks_init() -> *mut InstrumentHooks;
}
extern "C" {
    pub fn instrument_hooks_deinit(arg1: *mut InstrumentHooks);
}
extern "C" {
    pub fn instrument_hooks_is_instrumented(arg1: *mut InstrumentHooks) -> bool;
}
extern "C" {
    pub fn instrument_hooks_start_benchmark(arg1: *mut InstrumentHooks) -> u8;
}
extern "C" {
    pub fn instrument_hooks_stop_benchmark(arg1: *mut InstrumentHooks) -> u8;
}
extern "C" {
    pub fn instrument_hooks_set_executed_benchmark(
        arg1: *mut InstrumentHooks,
        pid: i32,
        uri: *const ::std::os::raw::c_char,
    ) -> u8;
}
extern "C" {
    pub fn instrument_hooks_executed_benchmark(
        arg1: *mut InstrumentHooks,
        pid: i32,
        uri: *const ::std::os::raw::c_char,
    ) -> u8;
}
extern "C" {
    pub fn instrument_hooks_set_integration(
        arg1: *mut InstrumentHooks,
        name: *const ::std::os::raw::c_char,
        version: *const ::std::os::raw::c_char,
    ) -> u8;
}
extern "C" {
    pub fn instrument_hooks_add_marker(
        arg1: *mut InstrumentHooks,
        pid: u32,
        marker_type: u8,
        timestamp: u64,
    ) -> u8;
}
extern "C" {
    pub fn instrument_hooks_current_timestamp() -> u64;
}
pub const instrument_hooks_feature_t_FEATURE_DISABLE_CALLGRIND_MARKERS: instrument_hooks_feature_t =
    0;
pub type instrument_hooks_feature_t = ::std::os::raw::c_uint;
extern "C" {
    pub fn instrument_hooks_set_feature(feature: instrument_hooks_feature_t, enabled: bool);
}
