pub const COMMENT: &[u8; 2] = b"--";
pub const VOID_TAGS: &[&[u8]; 15] = &[
    b"area", b"base", b"br", b"col", b"embed", b"hr", b"img", b"input", b"keygen", b"link",
    b"meta", b"param", b"source", b"track", b"wbr",
];
