// Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
// SPDX-License-Identifier: Apache-2.0 OR ISC
// Thu Mar 19 18:54:50 UTC 2026

pub(super) const CRYPTO_LIBRARY: &[&str] = &[
    "crypto/fipsmodule/ml_kem/mlkem/native/aarch64/src/intt.S",
    "crypto/fipsmodule/ml_kem/mlkem/native/aarch64/src/ntt.S",
    "crypto/fipsmodule/ml_kem/mlkem/native/aarch64/src/polyvec_basemul_acc_montgomery_cached_asm_k2.S",
    "crypto/fipsmodule/ml_kem/mlkem/native/aarch64/src/polyvec_basemul_acc_montgomery_cached_asm_k3.S",
    "crypto/fipsmodule/ml_kem/mlkem/native/aarch64/src/polyvec_basemul_acc_montgomery_cached_asm_k4.S",
    "crypto/fipsmodule/ml_kem/mlkem/native/aarch64/src/poly_mulcache_compute_asm.S",
    "crypto/fipsmodule/ml_kem/mlkem/native/aarch64/src/poly_reduce_asm.S",
    "crypto/fipsmodule/ml_kem/mlkem/native/aarch64/src/poly_tobytes_asm.S",
    "crypto/fipsmodule/ml_kem/mlkem/native/aarch64/src/poly_tomont_asm.S",
    "crypto/fipsmodule/ml_kem/mlkem/native/aarch64/src/rej_uniform_asm.S",
    "generated-src/ios-aarch64/crypto/chacha/chacha-armv8.S",
    "generated-src/ios-aarch64/crypto/cipher_extra/chacha20_poly1305_armv8.S",
    "generated-src/ios-aarch64/crypto/fipsmodule/aesv8-armx.S",
    "generated-src/ios-aarch64/crypto/fipsmodule/aesv8-gcm-armv8-unroll8.S",
    "generated-src/ios-aarch64/crypto/fipsmodule/aesv8-gcm-armv8.S",
    "generated-src/ios-aarch64/crypto/fipsmodule/armv8-mont.S",
    "generated-src/ios-aarch64/crypto/fipsmodule/bn-armv8.S",
    "generated-src/ios-aarch64/crypto/fipsmodule/ghash-neon-armv8.S",
    "generated-src/ios-aarch64/crypto/fipsmodule/ghashv8-armx.S",
    "generated-src/ios-aarch64/crypto/fipsmodule/keccak1600-armv8.S",
    "generated-src/ios-aarch64/crypto/fipsmodule/md5-armv8.S",
    "generated-src/ios-aarch64/crypto/fipsmodule/p256-armv8-asm.S",
    "generated-src/ios-aarch64/crypto/fipsmodule/p256_beeu-armv8-asm.S",
    "generated-src/ios-aarch64/crypto/fipsmodule/rndr-armv8.S",
    "generated-src/ios-aarch64/crypto/fipsmodule/sha1-armv8.S",
    "generated-src/ios-aarch64/crypto/fipsmodule/sha256-armv8.S",
    "generated-src/ios-aarch64/crypto/fipsmodule/sha512-armv8.S",
    "generated-src/ios-aarch64/crypto/fipsmodule/vpaes-armv8.S",
    "generated-src/ios-aarch64/crypto/test/trampoline-armv8.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/curve25519/bignum_madd_n25519.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/curve25519/bignum_madd_n25519_alt.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/curve25519/bignum_mod_n25519.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/curve25519/bignum_neg_p25519.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/curve25519/curve25519_x25519base_byte.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/curve25519/curve25519_x25519base_byte_alt.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/curve25519/curve25519_x25519_byte.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/curve25519/curve25519_x25519_byte_alt.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/curve25519/edwards25519_decode.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/curve25519/edwards25519_decode_alt.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/curve25519/edwards25519_encode.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/curve25519/edwards25519_scalarmulbase.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/curve25519/edwards25519_scalarmulbase_alt.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/curve25519/edwards25519_scalarmuldouble.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/curve25519/edwards25519_scalarmuldouble_alt.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/fastmul/bignum_emontredc_8n.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/fastmul/bignum_kmul_16_32.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/fastmul/bignum_kmul_32_64.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/fastmul/bignum_ksqr_16_32.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/fastmul/bignum_ksqr_32_64.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/generic/bignum_copy_row_from_table.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/generic/bignum_copy_row_from_table_16.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/generic/bignum_copy_row_from_table_32.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/generic/bignum_copy_row_from_table_8n.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/generic/bignum_ge.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/generic/bignum_mul.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/generic/bignum_optsub.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/generic/bignum_sqr.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p256/bignum_montinv_p256.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p256/p256_montjscalarmul.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p256/p256_montjscalarmul_alt.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p384/bignum_add_p384.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p384/bignum_deamont_p384.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p384/bignum_littleendian_6.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p384/bignum_montinv_p384.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p384/bignum_montmul_p384.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p384/bignum_montmul_p384_alt.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p384/bignum_montsqr_p384.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p384/bignum_montsqr_p384_alt.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p384/bignum_neg_p384.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p384/bignum_nonzero_6.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p384/bignum_sub_p384.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p384/bignum_tomont_p384.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p384/p384_montjdouble.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p384/p384_montjdouble_alt.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p384/p384_montjscalarmul.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p384/p384_montjscalarmul_alt.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p521/bignum_add_p521.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p521/bignum_fromlebytes_p521.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p521/bignum_inv_p521.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p521/bignum_mul_p521.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p521/bignum_mul_p521_alt.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p521/bignum_neg_p521.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p521/bignum_sqr_p521.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p521/bignum_sqr_p521_alt.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p521/bignum_sub_p521.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p521/bignum_tolebytes_p521.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p521/p521_jdouble.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p521/p521_jdouble_alt.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p521/p521_jscalarmul.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/p521/p521_jscalarmul_alt.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/sha3/sha3_keccak2_f1600.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/sha3/sha3_keccak4_f1600_alt.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/sha3/sha3_keccak4_f1600_alt2.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/sha3/sha3_keccak_f1600.S",
    "third_party/s2n-bignum/s2n-bignum-imported/arm/sha3/sha3_keccak_f1600_alt.S",
    "third_party/s2n-bignum/s2n-bignum-to-be-imported/arm/aes/aes-xts-dec.S",
    "third_party/s2n-bignum/s2n-bignum-to-be-imported/arm/aes/aes-xts-enc.S",
];
